@@ -1,10 +1,16 @@
-from typing import TYPE_CHECKING, List, Tuple
+from functools import wraps
+from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple
+
+from dvc_objects.fs.callbacks import DEFAULT_CALLBACK
 
 from ..hashfile.hash_info import HashInfo
 from ..hashfile.meta import Meta
 from ..hashfile.tree import Tree
 
 if TYPE_CHECKING:
+    from dvc_objects.fs.callbacks import Callback
+
+    from ..hashfile.db import HashFileDB
     from .index import BaseDataIndex, DataIndexKey
 
 
@@ -30,7 +36,7 @@ def md5(index: "BaseDataIndex") -> None:
             path = entry.path
 
         try:
-            meta = Meta.from_info(entry.fs.info(path))
+            meta = Meta.from_info(entry.fs.info(path), entry.fs.protocol)
         except FileNotFoundError:
             continue
 
@@ -64,7 +70,9 @@ def build_tree(
 
 
 def _save_dir_entry(
-    index: "BaseDataIndex", key: "DataIndexKey", odb=None
+    index: "BaseDataIndex",
+    key: "DataIndexKey",
+    odb: Optional["HashFileDB"] = None,
 ) -> None:
     from ..hashfile.db import add_update_tree
 
@@ -80,8 +88,26 @@ def _save_dir_entry(
     setattr(entry.meta, tree.hash_info.name, tree.hash_info.value)
 
 
-def save(index: "BaseDataIndex", odb=None, **kwargs) -> None:
+def _wrap_add(callback: "Callback", fn: Callable):
+    wrapped = callback.wrap_fn(fn)
+
+    @wraps(fn)
+    def func(path: str, *args, **kwargs):
+        kw: Dict[str, Any] = dict(kwargs)
+        with callback.branch(path, path, kw):
+            return wrapped(path, *args, **kw)
+
+    return func
+
+
+def save(
+    index: "BaseDataIndex",
+    odb: Optional["HashFileDB"] = None,
+    callback: "Callback" = DEFAULT_CALLBACK,
+    **kwargs
+) -> int:
     dir_entries: List["DataIndexKey"] = []
+    transferred = 0
 
     for key, entry in index.iteritems():
         assert entry.meta and entry.fs
@@ -103,12 +129,16 @@ def save(index: "BaseDataIndex", odb=None, **kwargs) -> None:
             cache = odb or entry.cache
             assert entry.hash_info.value
             assert cache
-            cache.add(
+            add = _wrap_add(callback, cache.add)
+            transferred += add(
                 path,
                 entry.fs,
                 entry.hash_info.value,
+                callback=callback,
                 **kwargs,
             )
 
     for key in dir_entries:
         _save_dir_entry(index, key, odb=odb)
+
+    return transferred
